#!/usr/bin/env python
# encoding: utf-8
"""
alignment_helpers.py

Created by Rod Docking on 2011-07-25.
Copyright (c) 2011 Canada's Michael Smith Genome Sciences Centre. All rights reserved.
"""

import sys
import os
import subprocess
from cluster_job import Cluster_job, submit_cluster_or_local
from cluster_job import hold_for_job_completion, remove_submission_files
from common_utils import validate_file

def align_all_reads(reference, reads_list, output_root, logger=None,
                    bam=False, cluster=False, cluster_opt=None):
    """
    Wrapper function to handle going from a set of unaligned reads to a set
        of aligned BAM files
    
    NOTE: If cluster is False, all the jobs will run locally - otherwise, they'll run as linked SGE
            cluster jobs
    """
    
    logger.debug("Starting alignment of reads in {0} to reference {1}".format(
        reads_list, reference))
    
    # Index the reference file in the main thread
    logger.debug("Indexing {0}".format(reference))
    C = Cluster_job(jobname = 'index_{0}.sh'.format(output_root))
    C.append(samtools_faidx(reference))
    C.append(bwa_index(reference))
    C.run_local()
    remove_submission_files([C])
    logger.debug("Finished indexing {0}".format(reference))
        
    # Infer read pairs - BAM files should be listed singly, fastq files in pairs
    read_files = [l.rstrip() for l in open(reads_list, 'r').readlines()]
    base_names = [os.path.basename(b) for b in read_files]
    
    # Because BAM files are listed singly, just go through the file list once
    if bam:
        base_names = [os.path.splitext(b)[0] for b in base_names]
        read_pairs = [(r,r) for r in read_files]
        base_name_pairs = [(b,b) for b in base_names]
        sai_pairs = [(b + '.1.sai', b + '.2.sai') for b in base_names]
        bam_files = ["{0}.{1}.bam".format(output_root, b) for b in base_names]
        sorted_bam_files = ["{0}.{1}.sorted.bam".format(output_root, b) for b in base_names]
        flag_files = ["{0}.{1}.flagstat".format(output_root, b) for b in base_names]

    # Fastq files are listed as pairs, so step through the list by two files at a time
    else:
        read_pairs, base_name_pairs, sai_pairs = [], [], []
        bam_files, sorted_bam_files, flag_files = [], [], []
        for i in range(0, len(read_files), 2):
            read_pairs.append((read_files[i], read_files[i+1]))
            base_name_pairs.append((base_names[i], base_names[i+1]))
            sai_pairs.append((base_names[i] + '.sai', base_names[i+1] + '.sai'))
            bam_files.append("{0}.{1}.bam".format(output_root, base_names[i]))
            sorted_bam_files.append("{0}.{1}.sorted.bam".format(output_root, base_names[i]))
            flag_files.append("{0}.{1}.flagstat".format(output_root, base_names[i]))

    # For each pair of read files:
    all_jobs_list, bam_list = [], []
    for read_pair, base_name_pair, sai_pair, bam_file, sorted_bam_file, flag_file in zip(
        read_pairs, base_name_pairs, sai_pairs, bam_files, sorted_bam_files, flag_files):
        
        # Submit two alignment jobs
        job_list = []
        for i in [0,1]:
            
            # Set up bwa aln job to use appropriate queue and resources
            C = Cluster_job(jobname='bwa_aln_{0}_{1}.sh'.format(
                            base_name_pair[i], i), touch=True,
                            host = cluster_opt.host,
                            pe=cluster_opt.bwa_aln_pe, 
                            cpus = cluster_opt.bwa_aln_threads,
                            queue = cluster_opt.bwa_aln_queue,
                            mem_free='4G', mem_token='4G', h_vmem = "4G",
                            )
            
            if bam:
                C.append(bwa_aln(reference, read_pair[i], sai_pair[i],
                        threads=cluster_opt.bwa_aln_threads, 
                        bam_read=str(i+1)))
            else:
                C.append(bwa_aln(reference, read_pair[i], sai_pair[i],
                        threads=cluster_opt.bwa_aln_threads, 
                        bam_read=None))

            # Submit alignment jobs and return job list
            job_list = submit_cluster_or_local(job_list, C, cluster)
        
        # Dependent bwa sampe job        
        C = Cluster_job(jobname='bwa_sampe_{0}.sh'.format(
                        base_name_pair[0]),
                        touch=True, host = cluster_opt.host,
                        mem_free='10G', mem_token='10G', h_vmem = "10G",
                        queue=cluster_opt.regular_queue,
                        dependencies=job_list)

        C.append(bwa_sampe(reference, sai_pair[0], sai_pair[1],
                  read_pair[0], read_pair[1], bam_file))

        # Sort, index, and flagstat
        C.append(samtools_sort(bam_file, sorted_bam_file[:-4]))
        C.append(samtools_index(sorted_bam_file))
        C.append(samtools_flagstat(sorted_bam_file, flag_file))

        # Submit jobs and add to job list
        job_list = submit_cluster_or_local(job_list, C, cluster)
        
        # Add these jobs onto the main tracking list
        all_jobs_list.extend(job_list)
        bam_list.append(sorted_bam_file)
    
    # Wait for all jobs to complete
    logger.info("Submitted {0} alignment jobs - waiting for completion".format(
        len(all_jobs_list)))
    hold_for_job_completion(all_jobs_list, sleep_timer=60, logger=logger)
    logger.info("Completed all alignment jobs")
        
    # Validate alignments here and remove intermediate files
    for sai_one, sai_two in sai_pairs:
        validate_file(sai_one, remove=True)
        validate_file(sai_two, remove=True)
    [validate_file(bam_file, remove=True) for bam_file in bam_files]
    [validate_file(sorted_bam_file) for sorted_bam_file in sorted_bam_files]
    [validate_file(flag_file, min_size=100) for flag_file in flag_files]
    
    # Clean up job submission files
    remove_submission_files(all_jobs_list)
    
    logger.debug("Finished alignment of reads in {0} to reference {1} - {2}".format(
        reads_list, reference, bam_list))
    logger.debug("Alignment statistics in {0}".format(flag_files))

    return bam_list

def bwa_index(reference):
    """
    Run bwa index for a given reference file
    """
    
    # Use different algorithm depending on file-size:
    ref_size = os.path.getsize(reference)
    if ref_size < 500000000:
        algo = 'is'
    else:
        algo = 'bwtsw'
    
    # Run bwa index to index reference file for BWA alignments
    return "bwa index -a {0} {1}".format(algo, reference)

def bwa_aln(reference, read_file, sai_file, threads='1', bam_read=None):
    """
    Run bwa aln to align reads in read_file against reference
    If sai_file is defined, write to that file - otherwise return stdout
    """
    
    # Run bwa aln from a bam file
    if bam_read:
        assert(bam_read in ['0','1','2'])
        args = "bwa aln -t{0} -b{1} {2} {3} > {4}".format(
            threads, bam_read, reference, read_file, sai_file)
    
    # Or from fastq input
    else:
        args = "bwa aln -t{0} {1} {2} > {3}".format(
            threads, reference, read_file, sai_file)
    
    return args

def bwa_samse(reference, sai_file, read_file, output_sam):
    """
    Run bwa samse to generate SAM from .sai input
    """
    
    args = "bwa samse {0} {1} {2} > {3}".format(
        reference, sai_file, read_file, output_sam)
    
    return args

def bwa_sampe(reference, sai_one, sai_two, reads_one, reads_two, bam_file):
    """
    Run bwa sampe to generate a BAM file from paired .sai input
    """
    
    # Run bwa sampe without Smith-Waterman stage
    args = "bwa sampe -s -r \'@RG\tID:1\tLB:1\tSM:1\' {0} {1} {2} {3} {4} |".format(
        reference, sai_one, sai_two, reads_one, reads_two)
    args = args + " samtools view -bS - > {0}".format(bam_file)    
    
    return args

def samtools_faidx(reference):
    """
    Run samtools faidx for a given reference file
    """
    
    return "samtools faidx {0}".format(reference)

def samtools_flagstat(bam_file, flag_file):
    """
    Run samtools flagstat for a given reference file
    """
    
    return "samtools flagstat {0} > {1}".format(bam_file, flag_file)

def samtools_index(bam_file):
    """
    Run samtools index
    """

    return "samtools index {0}".format(bam_file)

def samtools_sort(bam_file, sorted_bam_root):
    """
    Convert SAM to BAM and sort
    """

    return "samtools sort {0} {1}".format(
        bam_file, sorted_bam_root)

def write_bam_list(bam_list, output_root):
    """
    Write the list of BAM files to a file (for mpileup)
    
    """

    # Set file name with output root
    bam_list_file = "{0}.bam_files.in".format(output_root)

    # Write each bam file to the list and close
    bam_list_handle = open(bam_list_file, 'w')
    [bam_list_handle.write("{0}\n".format(b)) for b in bam_list]
    bam_list_handle.close()

    return bam_list_file

def main():
	pass


if __name__ == '__main__':
	main()

