#!/usr/bin/env python
# encoding: utf-8
"""
anchor_assembly_runner.py

Created by Rod Docking on 2011-07-26.
Copyright (c) 2011 Canada's Michael Smith Genome Sciences Centre. All rights reserved.
"""

import sys
import os
import shutil
import logging
from glob import iglob

from sam_record import SAM_record, bam_open
from fasta_laundry import build_header_dict, build_seq_dict, read_fasta, ambiguate, descaffold
from cluster_job import Cluster_job, submit_cluster_or_local
from cluster_job import hold_for_job_completion, remove_submission_files

def build_job_list(assembly_file, gap_dict, min_extend, logger):
    """Build a list of assembly and extension jobs"""
    
    header_dict = build_header_dict(assembly_file)
    job_list = []
    
    # Infer contigs to extend and format job list
    for scaffold in header_dict:
        
        # Extend if contig is longer than threshold
        c_len = header_dict[scaffold]
        
        # Note - set gap coordinates to -1,-1  for front extensions
        #   to match how REGULAR gaps are annotated
        assemble_front = [-1,-1] if c_len >= min_extend else []
        assemble_back = [c_len,c_len] if c_len >= min_extend else []
        
        # Add internal assemblies to list
        scaffold_list = gap_dict[scaffold] if scaffold in gap_dict else []
        
        # Add list of jobs to job list
        if assemble_front or assemble_back or scaffold_list:
             job_list.append((scaffold, assemble_front, scaffold_list, assemble_back))
    
    # Count jobs for log
    num_extensions, num_closures = 0, 0
    for j in job_list:
        if j[1]:
            num_extensions += 1
        num_closures += len(j[2])    
        if j[3]:
            num_extensions += 1

    logger.info("Attempting {0} contig extensions (threshold is {1}bp)".format(
        num_extensions, min_extend))
    logger.info("Attempting {0} gap closure assemblies".format(num_closures))    

    # Calculate an appropriate number of scaffolds to assemble per job
    # If it's less than 10, use a single job
    # Up to 1000, ten per job
    # Above 1000, limit to 100 jobs
    total_jobs = num_extensions + num_closures
    if total_jobs <= 10:
        scaffolds_per_job = 1000
    elif total_jobs <= 1000:
        scaffolds_per_job = 10
    else:
        scaffolds_per_job = total_jobs / 100        
    logger.info("{0} assemblies in total - splitting to {1} assemblies per cluster job".format(
        total_jobs, scaffolds_per_job))
        
    return job_list, scaffolds_per_job

def collect_sequences(original_assembly, job_list, output_root, min_gap, logger):
    """
    Find all the newly assembled sequences and write the new reference
    """
    
    anchor_seq_dict, novel_seq_dict = {}, {}
    novel_seq_file = "{0}.novel.fa".format(output_root)
    flank_seq_file = "{0}.flanks.fa".format(output_root)
    validation_seq_file = "{0}.validation.fa".format(output_root)
    anchored_assembly = "{0}.anchor.fa".format(output_root)
    all_results = "{0}.anchor.results.txt".format(output_root)
    
    num_gap_merge, num_gap_extend, num_gap_fail, num_end_extend, num_end_fail = 0, 0, 0, 0, 0
    
    # Re-write with iglob and shutil - NOTE:
    #   The old version of this function was taking quite a few hours when many
    #   assemblies were on the go - track this new one with lots of log messages
    
    # Get all the newly-assembled sequences
    logger.debug("Collecting newly-assembled sequences")
    novel_seq_handle = open(novel_seq_file, 'w')
    for novel_seq in iglob('scaffold_assemblies_{0}/sub_*/all_novel_seqs.fa'.format(output_root)):
        shutil.copyfileobj(open(novel_seq, 'r'), novel_seq_handle)
    novel_seq_handle.close()        
    logger.debug("Finished collecting newly-assembled sequences")

    # And all the sequences flanking scaffold gaps
    logger.debug("Collecting flank sequences")
    flank_seq_handle = open(flank_seq_file, 'w')
    for target_seq in iglob('scaffold_assemblies_{0}/sub_*/all_flank_seqs.fa'.format(output_root)):
        shutil.copyfileobj(open(target_seq, 'r'), flank_seq_handle)
    flank_seq_handle.close()
    logger.debug("Finished collecting flank sequences")
    
    # Descaffold the novel sequences, ambiguate both novel sequences and targets
    # (For validation analysis)
    validation_out = open(validation_seq_file, 'w')
    for F in read_fasta(flank_seq_file):
        F.long_seq = ambiguate(F.long_seq)
        validation_out.write(">{0}\n{1}\n".format(
            F.header_id, F.long_seq))
    
    for F in read_fasta(novel_seq_file):
        F.long_seq = ambiguate(F.long_seq)
        seq_list = descaffold(F.long_seq, min_gap)
        i = 0
        for seq in seq_list:
            validation_out.write(">{0}_{1}\n{2}\n".format(
                F.header_id, i, seq))
            i += 1
    validation_out.close()
            
    # All the concatenated results
    logger.debug("Collecting Anchor results")
    all_results_handle = open(all_results, 'w')
    for manifest_result in iglob('scaffold_assemblies_{0}/sub_*/job_manifest.txt.result'.format(output_root)):
        shutil.copyfileobj(open(manifest_result, 'r'), all_results_handle)
    all_results_handle.close()
    logger.debug("Finished collecting Anchor results")
    
    # Sum up the results for all jobs
    logger.debug("Collecting Anchor summaries")
    for single_result in iglob('scaffold_assemblies_{0}/sub_*/result.txt'.format(output_root)):
        c = open(single_result,'r').readlines()[0].rstrip().split()
        num_gap_merge += int(c[1])
        num_gap_extend += int(c[2])
        num_gap_fail += int(c[3])
        num_end_extend += int(c[4])
        num_end_fail += int(c[5])
    logger.debug("Finished collecting Anchor summaries")
    
    # Read the original assembly...
    logger.debug("Collecting reconstructed contigs")
    for final_seq in iglob('scaffold_assemblies_{0}/sub_*/*_final.fa'.format(output_root)):
        F = [F for F in read_fasta(final_seq)][0]
        anchor_seq_dict[F.header_id] = F
    logger.debug("Finished collecting reconstructed contigs")

    # ...  and insert any new sequences into the final assembly
    logger.debug("Writing new assembly")
    anchored_assembly_handle = open(anchored_assembly, 'w')
    for F in read_fasta(original_assembly):
        if F.header_id in anchor_seq_dict:
            A = anchor_seq_dict[F.header_id]
            anchored_assembly_handle.write(A.print_fasta())
        else:
            anchored_assembly_handle.write(F.print_fasta())    
    anchored_assembly_handle.close()
    logger.debug("Finished writing new assembly")
    
    logger.info("Anchor results - Gap merge: {0} Gap extend: {1} Gap fail: {2} End extend: {3} End fail: {4}".format(
        num_gap_merge, num_gap_extend, num_gap_fail, num_end_extend, num_end_fail))
    
    return anchored_assembly, novel_seq_file

def find_scaffold_locations(assembly_file, min_gap, logger):
    """
    Parse the sequence dictionary, and find scaffolded gaps larger than min_gap
    """
    
    # Initialize gap dictionary and counter for skipped gaps
    gap_dict = {}
    small_skip_counter = 0
    
    # For each contig in the assembly
    for F in read_fasta(assembly_file):
        
        seq = F.long_seq
        gap_dict[F.header_id] = []
        in_gap = False
        current_gap = [0, 0]
        
        # Go through the sequence and look for N characters
        for i in range(len(seq)):
            
            if seq[i] == 'N' and (not in_gap):
                in_gap = True
                current_gap[0] = i
            elif seq[i] == 'N' and (in_gap):
                continue
            elif seq[i] != 'N' and (in_gap):
                in_gap = False
                current_gap[1] = i-1
                
                # If the gap is long enough, add it to the dict
                gap_size = current_gap[1] - current_gap[0] + 1
                if gap_size >= min_gap: 
                    gap_dict[F.header_id].append(current_gap)
                else:
                    small_skip_counter += 1
                
                current_gap = [0, 0]
    
    # Go through the gap dictionary, and count gaps and bases in gaps
    num_gaps, num_contigs_with_gap, num_bases_in_gaps = 0, 0, 0
    for gap in gap_dict:
        
        if gap_dict[gap] == []:
            continue
        
        num_contigs_with_gap += 1
        for indiv_gap in gap_dict[gap]:
            num_gaps += 1
            num_bases_in_gaps += (indiv_gap[1]-indiv_gap[0]+1)
    
    logger.info("{0} distinct gaps in {1} distinct contigs, with {2} bases in gaps".format(
        num_gaps, num_contigs_with_gap, num_bases_in_gaps))
    logger.info("Skipped {0} gaps smaller than min_gap size of {1}".format(
        small_skip_counter, min_gap))

    return gap_dict

def infer_isize(bam_list_file):
    """
    From a BAM file, infer the average insert size and read length by
    looking at the first 100000 properly paired reads
    """
    
    result = []
    bam_list = open(bam_list_file,'r').readlines()
    bam_list = [b.rstrip() for b in bam_list]
    
    bam_handle = open(bam_list_file,'w')
    
    # For each bam file
    for bam_file in bam_list:
        
        read_length, isize_list = None, []
        for S in bam_open(bam_file):
            
            # Stop after 100000 reads
            if len(isize_list) >= 100000:
                break
            
            if S.header: continue
            
            # Check for variable read lengths
            if not read_length:
                read_length = S.read_length
            else:
                if S.read_length != read_length:
                    print "Warning! Unequal read lengths: ", read_length, S.read_length
            
            # Add properly paired read insert sizes to list
            if S.proper_pair() and (not S.isize == 0):
                isize_list.append(abs(S.isize))
        
        # Find the median insert size
        midpoint = len(isize_list)/2
        sorted_isize_list = sorted(isize_list)
        median_isize = sorted_isize_list[midpoint]
        
        # Append to result list
        result.append([bam_file,(read_length,median_isize)])
        bam_handle.write("{0}\t{1}\t{2}\n".format(os.path.abspath(bam_file), read_length, median_isize))
    
    bam_handle.close()
    
    return result

def scaffold_filler(sub_job, assembly_file, bam_list, anchor_opts):
    """ Format command for scaffold-filling job """
    
    filler_script = os.path.dirname(os.path.abspath(sys.argv[0])) + \
                    '/lib/anchor_scaffold_filler.py'
    return "{0} {1} -r{2} -o{3} -c{4} -n{5} {6} {7} job_manifest.txt result.txt".format(
            sys.executable, filler_script, anchor_opts['min_read'],
            anchor_opts['max_overlap'], anchor_opts['min_cov'],
            anchor_opts['min_gap'], assembly_file, bam_list)

def submit_assembly_job(sub_job, all_jobs_list, assembly_file, bam_list, 
                        anchor_opts, output_root, cluster=False, 
                        cluster_opt=None):
    """ Submit local assembly job """
    
    print "Submitting sub-job ", sub_job
    sub_job_dir = os.getcwd() + '/scaffold_assemblies_{0}/sub_{1}'.format(output_root, sub_job)
        
    C = Cluster_job(jobname="anchor_{0}.sh".format(sub_job),
                    directory=sub_job_dir,
                    touch=True, host=cluster_opt.host,
                    mem_free='4G', mem_token='4G',
                    h_vmem='4G', queue=cluster_opt.regular_queue)
                    
    C.append(scaffold_filler(sub_job, assembly_file, bam_list, anchor_opts))
    all_jobs_list = submit_cluster_or_local(all_jobs_list, C, cluster)
    
    return all_jobs_list

def run_anchor_assemblies(assembly_file, bam_list_file, output_root, logger, 
                            min_extend, max_overlap_proportion, min_cov, min_gap, cluster=False, cluster_opt=None):
    """ Infer assemblies to perform and run them """
    
    logger.debug("Starting Anchor local assemblies for alignments in {0} to reference {1}".format(
        bam_list_file, assembly_file))
        
    # Get read lengths and insert sizes from aligned BAM files
    isize_list = infer_isize(bam_list_file)
    logger.info("Inferred library read lengths and insert sizes:")
    logger.info(isize_list)
    
    # Set overlap at 60% of SHORTEST read length
    min_read_length = min([r[1][0] for r in isize_list])
    max_overlap = str(int(int(min_read_length) * max_overlap_proportion))
    anchor_opts = { 'min_read'      : min_read_length, 
                    'max_overlap'   : max_overlap,
                    'min_cov'       : min_cov, 
                    'min_gap'       : min_gap}
    logger.info("Anchor inferred parameters:")
    logger.info(anchor_opts)
        
    # Find all contigs with scaffolds in them, and generate a jobs list
    # DO NOT include gaps smaller than min_gap
    gap_dict = find_scaffold_locations(assembly_file, min_gap, logger)
        
    # Write out list of assembly jobs:
    job_list, scaffolds_per_job = build_job_list(assembly_file, gap_dict, min_extend, logger)
    
    # Set up main output directory
    scaffold_path = os.getcwd() + '/scaffold_assemblies_{0}'.format(output_root)
    if os.path.exists(scaffold_path):
        shutil.rmtree(scaffold_path)
    os.mkdir(scaffold_path)
    
    # Add full paths for input assembly and bam_list_file
    assembly_file = os.path.abspath(assembly_file)
    bam_list_file = os.path.abspath(bam_list_file)
    
    # Perform local assemblies
    i, job_handle, sub_job = 0, None, -1
    cluster_jobs_list = []
    for job in job_list:
        
        # If the current job list is filled
        if i >= scaffolds_per_job or not job_handle:
            
            # If there is a current job, submit it
            if job_handle:
                job_handle.close()
                cluster_jobs_list = submit_assembly_job(sub_job, cluster_jobs_list,
                                     assembly_file, bam_list_file, anchor_opts, output_root,
                                     cluster, cluster_opt)
            
            # Create the new sub-directory and job handle
            sub_job += 1
            i = 0
            sub_directory = scaffold_path + '/sub_{0}'.format(sub_job)
            os.mkdir(sub_directory)
            job_handle = open('{0}/job_manifest.txt'.format(sub_directory),'w')
        
        # Add scaffold to current job and increment the job counter
        job_handle.write("\t".join([str(k) for k in job]) + "\n")
        
        # Increment counter by the number of assemblies that need to be performed
        # Format of job list is: ID [assemble_front], [list_of_scaffolds], [assemble_back]
        if job[1]:
            i += 1
        if job[3]:
            i += 1
        i += len(job[2])
        
    # Submit last assembly job
    job_handle.close()
    cluster_jobs_list = submit_assembly_job(sub_job, cluster_jobs_list,
                         assembly_file, bam_list_file, anchor_opts, 
                         output_root, cluster, cluster_opt)

    
    # Wait for assembly jobs to finish
    hold_for_job_completion(cluster_jobs_list, sleep_timer=60, logger=logger)
    
    # Collect new and novel sequences
    anchored_assembly_file, novel_seq_file = collect_sequences(assembly_file,
                                              job_list, output_root, min_gap,
                                              logger)

    # Clean up scaffold assemblies directory
    #shutil.rmtree(scaffold_path)
    
    logger.debug("Finished Anchor local assemblies - returning {0} and {1}".format(
      anchored_assembly_file, novel_seq_file))
    
    return anchored_assembly_file, novel_seq_file

def main():

    # Set up logger
    logger = logging.getLogger("anchor_assembly_runner")
    logger.setLevel(logging.DEBUG)
    handler = logging.FileHandler("{0}.log".format(sys.argv[1]), 'a')
    formatter = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    
    # For now, just count gaps in input assembly:
    gap_dict = find_scaffold_locations(sys.argv[1], logger)
    print gap_dict

if __name__ == '__main__':
	main()

