#!/usr/bin/env python
# encoding: utf-8
"""
anchor_snp_indel_corrector.py

Created by Rod Docking on 2010-04-27.
Copyright (c) 2010 Canada's Michael Smith Genome Sciences Centre. All rights reserved.
"""

import sys
import os
import re
import string
import logging
from optparse import OptionParser
from vcf_record import VCF_record
from fasta_laundry import build_header_dict, read_fasta, FastaRecord

def parse_options():
    """Parse options provided on the command line"""
    
    usage = "usage: %prog [options] snps.mpileup.vcf reference.fa output.fa"
    parser = OptionParser(usage)
    
    parser.add_option("-o", "--hom",
        action="store_true", dest="hom",
        help="correct homozygous SNPs")
    parser.add_option("-i", "--indel",
        action="store_true", dest="indel",
        help="correct indels")
    
    (options, args) = parser.parse_args()
    pathed_args = [os.path.abspath(arg) for arg in args]
    
    if not (len(args) == 3):
        parser.error("incorrect number of arguments")
    
    return options, pathed_args

def parse_VCF(VCF_file, logger, correct_hom=True, correct_indel=True):
    """Parse VCF input to find SNPs and indels to correct"""
    
    correction_dict = {}
    heterozygous_counter, homozygous_counter, indel_counter, skipped_indel_counter, keep_ref_counter = 0, 0, 0, 0, 0
    
    for line in open(VCF_file, 'r'):
        
        V = VCF_record(line)
        correction_string = ''
        if V.header:
            continue
        
        # samtools sometimes calls for an insert of N, which we
        #   don't want to make
        if ("N" in V.alt.upper()) or ("N" in V.ref.upper()):
            continue
        
        if V.is_SNP():
            
            if V.is_homozygous() and correct_hom:
                homozygous_counter += 1
                correction_string = "{0} {1} {2}".format(V.pos, V.ref, V.alt)
            elif V.is_heterozygous():
                heterozygous_counter += 1
        
        elif V.is_indel():
            
            if correct_indel:
                if V.is_homozygous():
                    indel_counter += 1
                    correction_string = "{0} {1} {2}".format(V.pos, V.ref, V.alt)
                elif V.is_heterozygous():
                    skipped_indel_counter += 1
        else:
            print V.raw
            print "Not a SNP or an indel???"
            sys.exit()
        
        # If we've found a correction to make, add it to the dictionary
        if correction_string:
            
            if not correction_dict.has_key(V.chrom):
                correction_dict[V.chrom] = []
                correction_dict[V.chrom].append(correction_string)
            else:
                correction_dict[V.chrom].append(correction_string)
    
    logger.info("Correcting {0} homozygous positions, and {1} indel positions".format(
        homozygous_counter, indel_counter))
    
    return correction_dict

def correct_sequence(contig_to_correct, seq_to_correct, header_dict, correction_dict):
    """Correct SNPs in the passed sequence"""
    
    # Check that the length matches what we expect
    if not ( len(seq_to_correct) == int(header_dict[contig_to_correct]) ):
        print "Error! The seq length we found doesn't match the header dictionary!"
        sys.exit()
    
    # Need to keep track of the offset that we introduce by adding/removing bases
    #  Initially set to -1: string index is 0-based, pos coordinates from pileup are 1-based
    offset = -1
    lookahead = 0
    
    # For each base to correct:
    for base_to_correct in correction_dict[contig_to_correct]:
        cols = base_to_correct.split()
        position, ref_base, called_base = int(cols[0]), cols[1], cols[2]
        
        # If this SNP has been removed already because of a preceding indel, skip
        if position <= lookahead:
            sys.stderr.write("Skipping because of prior indel: {0}\n".format(cols))
            continue
        
        # For cases where there are multiple alternate alleles
        if "," in called_base:
            called_base = called_base[0]
        
        # Split SNP-correcting and indel-correcting logic:
        if len(called_base) == 1 and len(ref_base) == 1:
            # print "Ref: {0} Call: {1} Position: {2} Base at position: {3}".format(
            #     ref_base, called_base, position, seq_to_correct[position-1])
            
            # Check that it currently has the expected reference... note that the string index is 0-based, the position is 1-based
            try:
                assert(ref_base.upper() == seq_to_correct[position+offset].upper())
            except AssertionError:
                print "Error! The reference at the indicated position doesn't match the expected reference from the pileup file!"
                print contig_to_correct
                print seq_to_correct
                print correction_dict[contig_to_correct]
                print base_to_correct
                sys.exit()
            
            # Return a new string with the base corrected
            corrected_seq = seq_to_correct[:position + offset] + called_base + seq_to_correct[position + offset + 1:]
            
            # Check the the length hasn't changed
            if not(len(corrected_seq) == len(seq_to_correct)):
                print "Error! Old and new sequences are not the same length!"
                sys.exit()
            
            # Update for the next round
            seq_to_correct = corrected_seq
            
            # Check that the new string has the right base:
            try:
                assert(called_base.upper() == seq_to_correct[position+offset].upper())
            except AssertionError:
                print "Error! The reference at the indicated position doesn't match the expected reference from the pileup file!"
                print contig_to_correct
                print seq_to_correct
                print correction_dict[contig_to_correct]
                print base_to_correct
                sys.exit()
        else:
            
            # Find the ref base in the sequence
            ref_slice = seq_to_correct[position+offset:position+offset+len(ref_base)]
            assert(ref_slice == ref_base)
            
            # Correct the sequence
            corrected_seq = seq_to_correct[:(position+offset)] + called_base + seq_to_correct[position+offset+len(ref_base):]
            assert(len(corrected_seq) == ( len(seq_to_correct) + len(called_base) - len(ref_base) ) )
            
            # Update the sequence and offset
            seq_to_correct = corrected_seq
            offset += (len(called_base) - len(ref_base))
            
            # Make sure we don't try to correct any SNPs in positions that were removed
            lookahead = position + (len(ref_base) - len(called_base))
                
    return seq_to_correct

def correct_assembly_file(header_dict, correction_dict, fasta_file, output_file):
    """Parse the assembly fasta file, returning new contigs"""
    
    fasta_out = open(output_file, 'w')
    
    for F in read_fasta(fasta_file):
        
        # Quickly check that we have this contig in our reference
        assert(F.header_id in header_dict)
        
        # If it's one we need to correct, make the correction
        if F.header_id in correction_dict:
            F.long_seq = correct_sequence(F.header_id, F.long_seq, header_dict, correction_dict)
        
        # Write the original or modified sequence
        fasta_out.write(F.print_fasta())
    
    fasta_out.close()
    return

def scrub_assembly(vcf_file, reference_file, output_file, logger, correct_hom=True, correct_indel=True):
    """ Run the scrub pipeline """
    
    logger.debug("Starting assembly scrubbing for {0}, based on {1}".format(
        reference_file, vcf_file))
        
    # Parse the header dictionary
    header_dict = build_header_dict(reference_file)
    logger.info("Parsed header dictionary, found {0} sequences".format(
        len(header_dict.keys())))
    
    # Parse a VCF file
    correction_dict = parse_VCF(vcf_file, logger, correct_hom, correct_indel)
    logger.info("Parsed VCF file, found {0} contigs in need of correction".format(
        len(correction_dict.keys())))
    
    # Parse and correct the assembly fasta file
    correct_assembly_file(header_dict, correction_dict, reference_file, output_file)

    logger.debug("Finished assembly scrubbing - wrote {0}".format(
        output_file))

def main():
    """Correct a fasta file based on pileup-formatted SNP calls"""
    
    # Parse options
    opt, args = parse_options()
    vcf_file, reference_file, output_file = args
    
    # Set up logger
    logger = logging.getLogger("anchor_snp_indel_corrector")
    logger.setLevel(logging.DEBUG)
    handler = logging.FileHandler("{0}.log".format(output_file), 'a')
    formatter = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    
    # Log options and arguments
    logger.debug("Options: {0}".format(opt))
    logger.debug("Arguments: {0}".format(args))
        
    # Scrub assembly
    scrub_assembly(vcf_file, reference_file, output_file, 
                    logger, opt.hom, opt.indel)


if __name__ == '__main__':
    main()

