#!/usr/bin/env python
# encoding: utf-8
"""
assess_assembly.py

Created by Rod Docking on 2011-08-04.
Copyright (c) 2011 Canada's Michael Smith Genome Sciences Centre. All rights reserved.
"""

import sys
import os

from fasta_laundry import read_fasta, FastaRecord, reverse_complement, get_single_seq, descaffold
from psl_record import read_psl, psl_record

def assembly_check_front_target(blat_name, assembled_contig, front_target):
    """
    Check an assembly against the front target
    """
    
    # Read the blat file and extract the first hit
    hit_list = [P for P in read_psl(blat_name)]
    P = hit_list[0] if len(hit_list) > 0 else None
    if P: print P

    # Check for an appropriate hit - otherwise fail
    if (P) and (P.strand == '+') and (P.t_end == P.t_size) and \
        (P.num_mismatch <= 3) and (P.blockcount == 1) and (P.q_start == 0) and \
        ((P.q_size - P.q_end) > 1):
    
        # Extract the extension sequence
        start = P.q_end
        extension_seq = assembled_contig[start:]
        print "\tExtension by {0}bp,".format(P.q_size - P.q_end),
        print " starting from base {0} on the query sequence".format(start)
        print "\t", extension_seq, len(extension_seq)
        
        # Validate against target
        val_assem = assembled_contig[start-10:start]
        val_targ = front_target[-10:]
        if not (val_assem == val_targ):
            print "Warning! Potential mismatch:"
            print val_assem
            print val_targ
        
    else:
        print "\t==> No extension"
        extension_seq, start = '', None
        
    return extension_seq, P    

def assembly_check_back_target(blat_name, assembled_contig, back_target):
    """
    Check an assembly against the back target
    """
    
    # Read the blat file and extract the first hit
    hit_list = [P for P in read_psl(blat_name)]
    P = hit_list[0] if len(hit_list) > 0 else None
    if P: print P

    # Check for an appropriate hit - otherwise fail
    if (P) and (P.strand == '+') and (P.t_start == 0) and \
        (P.num_mismatch <= 3) and (P.blockcount == 1) and \
        (P.q_end == P.q_size) and (P.q_start > 1):
    
        # Extract the extension sequence and orient properly
        stop = P.q_start
        print "\tExtension by {0}bp, ".format(stop),
        print " starting from base {0} on the query sequence".format(stop)
        extension_seq = assembled_contig[:stop]
        print "\t", extension_seq, len(extension_seq)
        
        # Validate against target
        val_assem = assembled_contig[stop:stop+10]
        val_targ =  back_target[:10]
        if not (val_assem == val_targ):
            print "Warning! Potential mismatch:"
            print val_assem
            print val_targ
        
    else:
        print "\t==> No extension"
        extension_seq, start = '', None

    return extension_seq, P

def check_scaffold_assembly(key, gap, new_contig_seq, subcontig_list, result_dict, 
                            counter, forward=True, reverse=True):
    """Check a scaffold assembly for completeness"""
    
    # File names
    assembly_file_root = "{0}_{1}".format(key, counter)
    
    # Assemblies and targets
    assembly_file_fwd = assembly_file_root + "_assembled_fwd.fa"
    assembly_file_rev = assembly_file_root + "_assembled_rev.fa"
    front_flank_name = assembly_file_root + '_front_target.fa'
    back_flank_name = assembly_file_root + '_back_target.fa'
    
    # Blat files
    fwd_front_blat_name = assembly_file_root + '_fwd_front.psl'
    fwd_back_blat_name = assembly_file_root + '_fwd_back.psl'
    rev_front_blat_name = assembly_file_root + '_rev_front.psl'
    rev_back_blat_name = assembly_file_root + '_rev_back.psl'
    
    ####
    # Retrieve assembled contigs and targets
    ###
    
    if forward:
        forward_assembled_contig = get_single_seq(assembly_file_fwd)
        front_target = get_single_seq(front_flank_name)

    if reverse:
        reverse_assembled_contig = get_single_seq(assembly_file_rev)
        back_target = get_single_seq(back_flank_name)

    ###
    # Check all relevant hits for extensions
    ###

    # Check forward assembly extending from front target
    if forward:
        fwd_front_seq, fwd_front_hit = assembly_check_front_target(
            fwd_front_blat_name, forward_assembled_contig, front_target)
        
    # Check reverse assembly extending from back target
    if reverse:
        rev_back_seq, rev_back_hit = assembly_check_back_target(
            rev_back_blat_name, reverse_assembled_contig, back_target)
        
    # Check forward assembly to back target and reverse assembly to front target
    if forward and reverse:
        fwd_back_seq, fwd_back_hit = assembly_check_back_target(
            fwd_back_blat_name, forward_assembled_contig, back_target)
        rev_front_seq, rev_front_hit = assembly_check_front_target(
            rev_front_blat_name, reverse_assembled_contig, front_target)
       
    ###
    #   Extend / merge growing contig sequence
    ###   

    # Reverse assembly extending from the beginning of a contig - i.e. a back target
    if reverse and not forward:
        
        if rev_back_seq and rev_back_hit and back_target:
            print "\n\tReverse assembly extension: {0} {1}bp".format(
              rev_back_seq, len(rev_back_seq))
            
            # Check for coordinate mixups in targets vs. next contig
            #assert(back_target[:100] == new_contig_seq[:100])
            if not (back_target[:100] == new_contig_seq[:100]):
                print "Warning! Potential mismatch!"
                print back_target[:100]
                print new_contig_seq[:100]
            
            flank_sequence = rev_back_seq + new_contig_seq[:200]
            new_contig_seq = rev_back_seq + new_contig_seq
            result_dict['num_end_extend'] += 1
            output_string = "extend {0}".format(len(rev_back_seq))    

        else:
            print "\t==> No extension"
            flank_sequence = ''
            result_dict['num_end_fail'] += 1
            output_string = "extension fail"
    
    # Forward assembly extending from the end of a contig - i.e. a front target    
    elif forward and not reverse:
        
        if fwd_front_seq and fwd_front_hit and front_target:
            print "\n\tForward assembly extension: {0} {1}bp".format(
              fwd_front_seq, len(fwd_front_seq))

            # Check for coordinate mixups in targets vs. next contig
            #assert(front_target[-100:] == new_contig_seq[-100:])
            if not (front_target[-100:] == new_contig_seq[-100:]):
                print "Warning! Potential mismatch!"
                print front_target[-100:]
                print new_contig_seq[-100:]

            flank_sequence = new_contig_seq[-200:] + fwd_front_seq
            new_contig_seq = new_contig_seq + fwd_front_seq 
            result_dict['num_end_extend'] += 1
            output_string = "extend {0}".format(len(fwd_front_seq))    

        else:
            print "\t==> No extension"
            flank_sequence = ''
            result_dict['num_end_fail'] += 1
            output_string = "extension fail"
    
    # Gap-closing assembly - both forward and reverse assemblies attempted            
    else:
        
        # Calculate gap size and extract next contig
        original_gap_size = gap[1]-gap[0]+1
        next_contig_seq = subcontig_list.pop()
        
        # Check for coordinate mixups in targets vs. next contig
        #assert(front_target[-100:] == descaffold(new_contig_seq)[-1][-100:])
        #assert(back_target[:100] == descaffold(next_contig_seq)[0][:100])
        try:
            if not ( (front_target[-100:] == descaffold(new_contig_seq)[-1][-100:]) and 
                (back_target[:100] == descaffold(next_contig_seq)[0][:100]) ):
                print "Warning! Potential mismatch!"
                print front_target[-100:]
                print descaffold(new_contig_seq)[-1][-100:]
                print back_target[:100]
                print descaffold(next_contig_seq)[0][:100]
        except TypeError, e:
            print "Error examining flank sequences"
            print front_target
            print back_target
            print new_contig_seq
            print next_contig_seq
                            
        # check for forward assembly closure
        if fwd_front_seq and fwd_back_seq:

            print "\tForward assembly closes gap"
            new_contig_seq, flank_sequence = close_gap(
                fwd_front_seq, fwd_front_hit, fwd_back_hit, new_contig_seq, next_contig_seq)
            result_dict['num_gap_merge'] += 1
            output_string = "gap close forward"

        # check for reverse assembly closure
        elif rev_back_seq and rev_front_seq:
            
            print "\tReverse assembly closes gap"
            new_contig_seq, flank_sequence = close_gap(
                rev_front_seq, rev_front_hit, rev_back_hit, new_contig_seq, next_contig_seq)
            result_dict['num_gap_merge'] += 1
            output_string = "gap close reverse"
                    
        # fill in from edges
        elif fwd_front_seq or rev_back_seq:
            print "\tExtending into scaffold gap"
            
            # Calculate the remaining pad size to leave
            init_pad_size = original_gap_size - len(fwd_front_seq) - len(rev_back_seq)
            pad_size = 50 if init_pad_size < 25 else init_pad_size

            # Construct scaffold sequence from extensions and pads
            pad_sequence = 'N' * pad_size
            scaffold_seq = fwd_front_seq + pad_sequence + rev_back_seq
            
            # Insert scaffold sequence into growing contig
            flank_sequence = new_contig_seq[-200:] + scaffold_seq + next_contig_seq[:200]
            new_contig_seq = new_contig_seq + scaffold_seq + next_contig_seq
            result_dict['num_gap_extend'] += 1
            output_string = "gap reduced"
            
        # Otherwise fail
        else:
            print "\t==> No extension"
            
            # Re-insert the original gap
            scaffold_seq = 'N' * original_gap_size
            flank_sequence = ''
            new_contig_seq = new_contig_seq + scaffold_seq + next_contig_seq
            result_dict['num_gap_fail'] += 1
            output_string = "gap fail"
                
    return new_contig_seq, result_dict, output_string, flank_sequence

def close_gap(seq, front_hit, back_hit, new_contig_seq, next_contig_seq):
    """
    Close a gap and return the merged contig
    """

    # Insert this portion of the assembled sequence between the two edges
    if front_hit.q_end <= back_hit.q_start:
        
        # The extension sequence is already trimmed at the front at this point
        scaff_size = back_hit.q_start - front_hit.q_end
        scaffold_seq = seq[:scaff_size]
        
        flank_sequence = new_contig_seq[-200:] + scaffold_seq + next_contig_seq[:200]
        new_contig_seq = new_contig_seq + scaffold_seq + next_contig_seq

    # Blat hits imply that the two edges already overlap
    else:
        print "\tOverlap!!"
        overlap = front_hit.q_end - back_hit.q_start
        
        print new_contig_seq[-overlap:]
        print next_contig_seq[:overlap]
        if not new_contig_seq[-overlap:] == next_contig_seq[:overlap]:
            print "Warning!! Overlap sequences don't match"
        
        flank_sequence = new_contig_seq[-200:] + next_contig_seq[overlap:overlap + 200]
        new_contig_seq = new_contig_seq + next_contig_seq[overlap:]

    return new_contig_seq, flank_sequence

def main():
    pass


if __name__ == '__main__':
    main()

