#!/usr/bin/env python
# encoding: utf-8
"""
cluster_job.py

Written by Rod Docking (rdocking@bcgsc.ca)
"""

import sys
import os
import subprocess
from time import sleep
from optparse import OptionParser
from glob import glob
import socket
import re

class submissionError(Exception):
    """Simple exception class for parsing errors"""
    def __init__(self, message):
        self.message = message
        print "Job submission failed!!: {0}".format(message)

def parse_options(default_job):
    """Parse options provided on the command line"""
    
    usage = "usage: %prog [options] \'commands_to_run\'"
    parser = OptionParser(usage)
    
    parser.add_option("-e", "--host",
        action="store", dest="host", type="string",
        help="Cluster head node name [{0}]".format(default_job.host),
        default=default_job.host)
    parser.add_option("-d", "--dir",
        action="store", dest="directory", type="string",
        help="Starting directory [{0}]".format(default_job.directory),
        default=default_job.directory)
    parser.add_option("-j", "--jobname",
        action="store", dest="jobname", type="string",
        help="Job name [{0}]".format(default_job.jobname),
        default=default_job.jobname)
    parser.add_option("-w", "--dependencies",
        action="store", dest="dependencies", type="string",
        help="Comma-separated list of dependencies [{0}]".format(default_job.dependencies),
        default=default_job.dependencies)
    parser.add_option("-q", "--queue",
        action="store", dest="queue", type="string",
        help="Job queue [{0}]".format(default_job.queue),
        default=default_job.queue)
    parser.add_option("-P", "--project",
        action="store", dest="project", type="string",
        help="Job project [{0}]".format(default_job.project),
        default=default_job.project)
    parser.add_option("-p", "--parallel",
        action="store", dest="pe", type="string",
        help="Parallel environment and cpus [{0}]".format(default_job.pe),
        default=default_job.pe)
    parser.add_option("-n", "--cpus",
        action="store", dest="cpus", type="string",
        help="Number of cpus to request [{0}]".format(default_job.cpus),
        default=default_job.cpus)
    parser.add_option("-m", "--mem_token",
        action="store", dest="mem_token", type="string",
        help="Memory reservation [{0}]".format(default_job.mem_token),
        default=default_job.mem_token)
    parser.add_option("-t", "--tmp_token",
        action="store", dest="tmp_token", type="string",
        help="Temp space reservation [{0}]".format(default_job.tmp_token),
        default=default_job.tmp_token)
    parser.add_option("-x", "--excl",
        action="store_true", dest="excl", 
        help="Exclusive reservation (excl=true)",
        default=default_job.excl)
    parser.add_option("-f", "--cmd_file",
        action="store", dest="cmd_file", type="string",
        help="File with commands to run",
        default=None)
    parser.add_option("-B", "--batch",
        action="store_true", dest="batch",
        help="Each line in CMD_FILE should be run as a separate job",
        default=None)
    parser.add_option("-M", "--mpi",
        action="store_true", dest="mpi",
        help="Reserve a whole mpi.q node (-q mpi.q -m 4G -p ncpus -n 12)",
        default=False)
    parser.add_option("-N", "--no_submit",
        action="store_true", dest="write_only",
        help="Generate job submission scripts, but don't submit the jobs",
        default=False)
    parser.add_option("-T", "--touch",
        action="store_true", dest="touch",
        help="Touch a completion file to indicate job completion",
        default=False)
    
    (opt, args) = parser.parse_args()
    
    # Convenience option for mpi nodes
    if opt.mpi:
        opt.queue = "mpi.q"
        opt.pe = "ncpus"
        opt.cpus = "12"
        opt.excl = True
    else:
        opt.mem_free = opt.mem_token
        opt.h_vmem = opt.mem_token
    
    return opt, args

def hold_for_job_completion(job_list, sleep_timer=60, max_iterations=10000, logger=None):
    """
    Wait until all jobs are complete - input is a list of ClusterJob objects
    """
    
    # Filter out entries from local jobs
    job_list = [C for C in job_list if not C.job_id == 0]
    sys.stderr.write("Waiting for {0} jobs:\n".format(len(job_list)))

    i = 0
    while i < max_iterations:
    
        # For each job, check for touch file:
        num_complete, num_remaining, first_incomplete = 0, 0, None
        for j in range(len(job_list)):
        
            C = job_list[j]
            touch_file = C.pathed_jobname + ".complete"
            if os.path.exists(touch_file):
                num_complete += 1
            else:
                num_remaining += 1
                if not first_incomplete:
                    first_incomplete = C    
        
        if num_remaining > 0:
            if logger:
                logger.debug("{0} jobs complete, {1} remaining, first incomplete is {2} (Job ID {3})".format(
                    num_complete, num_remaining, first_incomplete.jobname, first_incomplete.job_id))
            i += 1
            sleep(sleep_timer)
        else:
            break
    
    logger.debug("Finished waiting for cluster jobs")    
    return

def remove_submission_files(job_list):
    """
    Remove submission and output files for all jobs in list
    """
    
    for C in job_list:
        files = ["{0}.{1}{2}".format(C.pathed_jobname, ext, C.job_id) for ext
                    in ['o', 'e', 'po', 'pe']] + [C.pathed_jobname,
                    C.pathed_jobname + ".complete"]
        for f in files:
            if os.path.exists(f):
                os.unlink(f)
    
    return

def submit_cluster_or_local(job_list, C, cluster=False):
    """
    Either run a job locally, or submit it to the cluster and add a dependency
    """
    
    # Run the job either locally or on the cluster
    if cluster:
        C.submit()
    else:
        C.run_local()
    
    # Add the job to be submitted to the list of dependent jobs
    job_list.append(C)
    
    return job_list

class Cluster_job:
    """A class for creating and submitting cluster jobs"""
    
    def __init__(self, host = None, jobname = (os.environ.get("USER") + '_job.sh'), directory = os.getcwd(),
                touch = False, dependencies = None, queue = None, project = None, pe = None, cpus = None,
                mem_token = "4G", mem_free = "4G", h_vmem = "4G", tmp_token = None, excl = False):
        """Create the header code for the submission script"""
        
        # Try to figure out which host to submit on
        self.host = host
        self.localhost = socket.gethostname()
        
        # If a head-node host isn't specified, try to infer the correct one
        if not self.host:
            if 'xhost' in self.localhost:
                self.host = 'apollo'
            elif 'login' in self.localhost:
                self.host = 'login4'
            elif 'node' in self.localhost:
                self.host = 'login4'
            else:
                self.host = 'apollo'
        
        # Get fully-pathed job name
        self.jobname = jobname
        self.directory = directory
        self.base_jobname = os.path.basename(jobname)
        self.pathed_jobname = os.path.join(directory, self.base_jobname)
        
        # If the job directory doesn't exist, create it
        if not os.path.exists(self.directory):
           os.makedirs(self.directory)
        
        # Set remaining class attributes
        self.touch = touch
        self.dependencies = dependencies
        self.queue = queue
        self.project = project
        self.pe = pe
        self.cpus = cpus
        self.mem_token = mem_token
        self.mem_free = mem_free
        self.h_vmem = h_vmem
        self.tmp_token = tmp_token
        self.excl = excl
        
        self.job_lines = []
    
    def append(self, cmd):
        """Append a command to the job"""
        
        self.job_lines.append("{0}".format(cmd))
    
    def submit(self, write_only = False):
        """Submit the job"""
        
        # Open the script file and add header info
        self.job_handle = open(self.pathed_jobname, 'w')
        self.job_handle.write('#! /bin/sh\n#$ -S /bin/sh\n#$ -N {0}\n#$ -V\n#$ -R y\n'.format(self.jobname))
        self.job_handle.write('#$ -wd {0}\n'.format(self.directory))
        
        # Split and re-join dependencies list to make sure it's comma-separated
        if self.dependencies:
            
            # Handle either a list of Cluster_job objects or..
            if isinstance(self.dependencies[0], Cluster_job):
                job_ids = [str(C.job_id) for C in self.dependencies]
                self.dependencies = ",".join(job_ids)
            # A string of job ids
            elif len(self.dependencies.split()) > 1:
                self.dependencies = ",".join(self.dependencies)
            self.job_handle.write("#$ -hold_jid {0}\n".format(self.dependencies))
        
        # Set queue
        if self.queue:
            self.job_handle.write("#$ -q {0}\n".format(self.queue))
        
        # Set project
        if self.project:
            self.job_handle.write("#$ -P {0}\n".format(self.project))
        
        # Set parallel environment options
        if self.pe and self.cpus:
            self.job_handle.write("#$ -pe {0} {1}\n".format(self.pe, self.cpus))
        
        # Set tmp space reservations
        if self.tmp_token:
            self.job_handle.write("#$ -l tmp_token={0}\n".format(self.tmp_token))
        
        # Set exclusive
        if self.excl:
            self.job_handle.write("#$ -l excl=true\n")
        
        # Set memory reservations
        # NOTE - we're setting these as hard limits here!
        self.job_handle.write("#$ -l mem_token={0}\n".format(self.mem_token))
        self.job_handle.write("#$ -l mem_free={0}\n".format(self.mem_free))
        self.job_handle.write("#$ -l h_vmem={0}\n\n\n".format(self.h_vmem))
                
        # Capture current path (from submission host) and cd to cwd
        self.job_handle.write("export PATH={0}\n".format(os.environ['PATH']))
        self.job_handle.write("cd {0}\n".format(self.directory))
        
        # Add job commands
        self.job_handle.write("\n".join(self.job_lines))
        
        # Add touch command if required
        if self.touch:
            self.job_handle.write("\ntouch {0}.complete\n".format(self.jobname))
        
        # Close off job file
        self.job_handle.write("\necho Job {0} complete!\n".format(self.jobname))
        self.job_handle.close()
        
        job_pattern = re.compile("^Your job (\d+)")
        
        # ssh to head node and submit job
        if not write_only:
            proc = subprocess.Popen(
                ['ssh', self.host, 'qsub', self.pathed_jobname],
                shell=False, cwd=os.getcwd(), stderr=subprocess.PIPE,
                stdout=subprocess.PIPE,
                )
            proc.wait()
            
            # Grab output from cluster
            self.submission_lines_stdout =  proc.stdout.readlines()
            self.submission_lines_stderr = proc.stderr.readlines()
            
            # Check for appropriate response, otherwise raise an error
            if len(self.submission_lines_stdout) == 1:
                match = re.match(job_pattern, self.submission_lines_stdout[0])
            else:
                raise submissionError(self.submission_lines_stderr)
            
            # Capture job id
            if match:
                self.job_id = match.groups()[0]
                print self.submission_lines_stdout[0],
            else:
                raise submissionError(self.submission_lines_stderr)
        
        sleep(1)
    
    def run_local(self):
        """ Run a job-submission file on the local host """
        
        # Create the job submission file and give it a fake job ID
        self.submit(write_only=True)
        self.job_id = 0
        
        # Run the job on the local host
        proc = subprocess.check_call(['sh', self.pathed_jobname], shell=False, cwd=os.getcwd())


if __name__ == "__main__":
    """ Self-test code for the Cluster_job class """
    
    # Create a dummy object to get the defaults (so we're not editing them in two places)
    #  If the user updates any options, those will be provided in place of the defaults
    default_job = Cluster_job()
    opt, args = parse_options(default_job)
    
    # Batch mode - create a job for each line in the command file (similar to mqsub)
    if opt.cmd_file and opt.batch:
        
        iterator = 0
        
        for line in open(opt.cmd_file, 'r'):
            # Create the job to submit
            C = Cluster_job(host = opt.host, jobname = opt.jobname + '_' + str(iterator),
                        directory = opt.directory, touch = opt.touch, dependencies = opt.dependencies,
                        queue = opt.queue, project = opt.project, pe = opt.pe, cpus = opt.cpus,
                        mem_token = opt.mem_token, mem_free = opt.mem_free, h_vmem = opt.h_vmem,
                        tmp_token = opt.tmp_token, excl=opt.excl)
            
            C.append(line.rstrip())
            C.submit(write_only = opt.write_only)
            iterator += 1
    
    # Regular mode - either submit a job with the command line arguments
    #  or with all the lines of the input file as a single job
    else:
        
        # Create the job to submit
        C = Cluster_job(host = opt.host, jobname = opt.jobname, 
            directory = opt.directory, touch = opt.touch,
            dependencies = opt.dependencies, queue = opt.queue, 
            project = opt.project, pe = opt.pe, cpus = opt.cpus,
            mem_token = opt.mem_token, mem_free = opt.mem_free, 
            h_vmem = opt.h_vmem, tmp_token = opt.tmp_token, 
            excl = opt.excl)
        
        if opt.cmd_file:
            for line in open(opt.cmd_file, 'r'):
                C.append(line.rstrip())
        
        else:
            C.append(args[0])
        
        # Submit the job
        C.submit(write_only = opt.write_only)
    
    
    