#!/usr/bin/env python
# encoding: utf-8
"""
psl_record.py

Created by Rod Docking on 2010-10-20.
Copyright (c) 2010 Canada's Michael Smith Genome Sciences Centre. All rights reserved.
"""

import sys
import os

def read_psl(psl_file):
    """ Generator function to read a psl file and yield psl_record objects"""
    
    # Return if the file doesn't exist
    if not os.path.exists(psl_file):
        raise StopIteration
    
    # Otherwise yield each new psl record
    with open(psl_file,'r') as psl_handle:
        for line in psl_handle:
            
            P = psl_record(line)
            if P.header_line:
                continue
            else:
                yield P

def longest_hit_per_query(psl_file):
    """ 
    Return the longest hit per query - 
    assumes query-sorted file (which is the default for blat)
    if there are two with equal length, return the first
    """

    current_query = None
    with open(psl_file,'r') as psl_handle:
        for line in psl_handle:

            P = psl_record(line)
            if P.header_line:
                continue
            
            # For the first record
            if not current_query:
                current_query = P
                current_query.num_hits = 1
            
            # New query    
            if not P.q_name == current_query.q_name:
                yield current_query
                current_query = P
                current_query.num_hits = 1
            
            # Same query
            else:
                if P.match_length > current_query.match_length:
                    tmp_hits = current_query.num_hits
                    current_query = P
                    current_query.num_hits = tmp_hits
                current_query.num_hits += 1    

    yield current_query


class pslError(Exception):
    """Simple exception class for parsing errors"""
    def __init__(self, message):
        self.message = message
        print "Parsing error: {0}".format(message)

class psl_record:
    def __init__(self, line):
        """Parse a single line of psl input"""
        
        c = line.split()
        
        # Return nothing for a header line
        try:
            int(c[0])
        except ValueError, e:
            self.header_line = True
            return
        except IndexError, e:
            self.header_line = True
            return
        
        # Check for the right number of columns
        if not len(c) == 21:
            raise pslError, "improper number of columns - expecting 21, found {0}\n{1}".format(len(c), line)
        
        self.header_line = False
        self.match_length = int(c[0])
        self.num_mismatch = int(c[1])
        self.num_rep_match = int(c[2])
        self.num_N = int(c[3])
        self.num_q_inserts = int(c[4])
        self.num_q_insert_bases = int(c[5])
        self.num_t_inserts = int(c[6])
        self.num_t_insert_bases =  int(c[7])
        self.strand = c[8]
        self.q_name = c[9]
        self.q_size = int(c[10])
        self.q_start = int(c[11])
        self.q_end = int(c[12])
        self.t_name = c[13]
        self.t_size = int(c[14])
        self.t_start = int(c[15])
        self.t_end = int(c[16])
        self.blockcount = int(c[17])
        self.blocksizes = c[18]
        self.q_starts = c[19]
        self.t_starts = c[20]
        
        # Convenience fields
        self.match_proportion = float(self.match_length) / float(self.q_size)
        self.raw = line
    
    def __repr__(self):
        return "<psl_record: {0} hits {1}>".format(self.q_name, self.t_name)
    
    def __str__(self):
        return "\n\t{0}bp hit from query {1} ({2}bp) to target {3} ({4}bp) \
            \n\t{5}-{6} on query to {7}-{8} on target {9} strand \
            \n\t{10} mismatch base(s) in {11} block(s)".format(
            self.match_length, self.q_name, self.q_size, self.t_name, self.t_size,
            self.q_start, self.q_end, self.t_start, self.t_end, self.strand,
            self.num_mismatch, self.blockcount)
    
    # Utility functions
    def is_full_length_hit(self):
        return True if self.match_length == self.q_size else False

    def block_zip(self):
        """
        Zip together the various block attributes
        """

        block_sizes_list = [int(i) for i in self.blocksizes.split(',')[:-1]]
        q_starts_list = [int(i) for i in self.q_starts.split(',')[:-1]]
        t_starts_list = [int(i) for i in self.t_starts.split(',')[:-1]]
        assert(len(block_sizes_list) == len(q_starts_list) == len(t_starts_list))
        
        print "q_start\tt_start"
        for block, q_start, t_start in zip(block_sizes_list, q_starts_list, t_starts_list):
            print "{0:,d}\t{1:,d}\t({2:,d}bp)".format(q_start, t_start, block)
        
        return     

    def full_report(self):
        """
        Very verbose output for a single psl record
        """

        print "{0:,d}bp hit from query {1} ({2:,d}bp) to target {3} ({4:,d}bp)".format(
            self.match_length, self.q_name, self.q_size, self.t_name, self.t_size)
        print "{0:,d}-{1:,d} on query to {2:,d}-{3:,d} on target {4} strand".format(
            self.q_start, self.q_end, self.t_start, self.t_end, self.strand)
        print "{0} mismatch base(s), {1} repeat base(s), {2} N base(s)".format(
            self.num_mismatch, self.num_rep_match, self.num_N)         
        print "Query: {0} inserts ({1:,d}bp)".format(self.num_q_inserts, 
            self.num_q_insert_bases)
        print "Target: {0} inserts ({1:,d}bp)".format(self.num_t_inserts, 
            self.num_t_insert_bases)
        self.block_zip()
        return

if __name__ == '__main__':
    """
    Print each record of a PSL file
    """
    
    #for P in read_psl(sys.argv[1]):
    #    print P
    for P in longest_hit_per_query(sys.argv[1]):
        print "{0}\t({1})".format(P.raw.rstrip(), P.num_hits)
    