#!/usr/bin/env python
# encoding: utf-8
"""
anchor_MS_analysis.py

Created by Rod Docking on 2011-09-18.
Copyright (c) 2011 Canada's Michael Smith Genome Sciences Centre. All rights reserved.
"""

import sys
import os

sys.path.append(os.path.join(os.path.dirname(sys.argv[0]), ".."))
from lib.fasta_laundry import reverse_complement, read_fasta, FastaRecord, ambiguate, descaffold
from lib.cluster_job import Cluster_job

def results_one(root_dir):
    """
    Results section 1 - S. cerevisiae test set
    """
    #   - Create results directory
    #   - Link in S. cerevisiae contig
    #   - Generate test data set
    #   - Run Anchor for 5 iterations
    #   - Validate results
    
    C = Cluster_job(jobname = 'anchor_MS_1.sh', directory = root_dir,
                    mem_token = "4G", mem_free = "4G", h_vmem = "4G")

    C.append("mkdir S_cerevisiae_test")
    C.append("cd S_cerevisiae_test")
    C.append("ln -s /home/rdocking/seqdev/Anchor/tests/S_cerevisiae_100kb.fa")
    C.append("python2.7 /home/rdocking/seqdev/Anchor/tests/generate_test_set.py S_cerevisiae_100kb.fa test")
    
    # Round 1
    C.append("python2.7 ~/seqdev/Anchor/anchor_pipeline.py -SAr -t50 -f200 -d3 -n25 test-contigs.fa test.fastq.in test_1")
    C.append("bwa bwasw /scratch/rdocking_prj/Anchor_MS/reference_sequences/S_cerevisiae_100kb.fa test_1.validation.fa | samtools calmd -bSu - S_cerevisiae_100kb.fa | samtools sort - test_1.validation")    
    C.append("samtools index test_1.validation.bam")
    C.append("python2.7 ~/seqdev/Anchor/tests/reference_validation.py test_1.anchor.results.txt test_1 test_1.validation.bam > test_1.validation.out")
    
    # Rounds 2-10
    for i in range(1,10):
        C.append("python2.7 ~/seqdev/Anchor/anchor_pipeline.py -SAr -t50 -f200 -d3 -n25 test_{0}.anchor.fa test.fastq.in test_{1}".format(i, i+1))
        C.append("bwa bwasw /scratch/rdocking_prj/Anchor_MS/reference_sequences/S_cerevisiae_100kb.fa test_{0}.validation.fa | samtools calmd -bSu - S_cerevisiae_100kb.fa | samtools sort - test_{0}.validation".format(i+1))    
        C.append("samtools index test_{0}.validation.bam".format(i+1))
        C.append("python2.7 ~/seqdev/Anchor/tests/reference_validation.py test_{0}.anchor.results.txt test_{0} test_{0}.validation.bam > test_{0}.validation.out".format(i+1))
    
    C.submit()
    
    return

def results_two(root_dir):
    """
    Results section 2 - Small genomes
    """
    
    #   - Create directories for Human BAC, E. coli
    #   - Submit Anchor jobs
    #   - Validate results
    
    C = Cluster_job(jobname = 'anchor_MS_2a.sh', directory = root_dir,
                    mem_token = "4G", mem_free = "4G", h_vmem = "4G")

    C.append("mkdir EC")
    C.append("cd EC")
    C.append("ln -s /scratch/rdocking_prj/Anchor_MS/input_assemblies/SRP000220-scaffolds.fa")
    C.append("ln -s /scratch/rdocking_prj/Anchor_MS/input_assemblies/SRP000220.in")
    C.append("python2.7 /home/rdocking/seqdev/Anchor/anchor_pipeline.py -SA -r -f200 -d3 -n25 -c genesis SRP000220-scaffolds.fa SRP000220.in ec_1")
    C.append("bwa bwasw /scratch/rdocking_prj/Anchor_MS/reference_sequences/NC_000913.fa ec_1.validation.fa | samtools calmd -bSu - /scratch/rdocking_prj/Anchor_MS/reference_sequences/NC_000913.fa | samtools sort - ec_1.validation")    
    C.append("samtools index ec_1.validation.bam")
    C.append("python2.7 ~/seqdev/Anchor/tests/reference_validation.py ec_1.anchor.results.txt ec_1 ec_1.validation.bam > ec_1.validation.out")
    for i in range(1,5):
        C.append("python2.7 /home/rdocking/seqdev/Anchor/anchor_pipeline.py -SA -r -f200 -d3 -n25 -c genesis ec_{0}.anchor.fa SRP000220.in ec_{1}".format(i, i+1))
    C.submit()

    D = Cluster_job(jobname = 'anchor_MS_2b.sh', directory = root_dir,
                    mem_token = "4G", mem_free = "4G", h_vmem = "4G")

    D.append("mkdir HS_bac")
    D.append("cd HS_bac")
    D.append("ln -s /scratch/rdocking_prj/Anchor_MS/input_assemblies/HTa14_0292A13-6.fa")
    D.append("ln -s /scratch/rdocking_prj/Anchor_MS/input_assemblies/HS0674.in")
    D.append("python2.7 /home/rdocking/seqdev/Anchor/anchor_pipeline.py -SA -r -f200 -d3 -n25 -c genesis HTa14_0292A13-6.fa HS0674.in bac_1")
    D.append("bwa bwasw /scratch/rdocking_prj/Anchor_MS/reference_sequences/human.fasta bac_1.validation.fa | samtools calmd -bSu - /scratch/rdocking_prj/Anchor_MS/reference_sequences/human.fasta | samtools sort - bac_1.validation")    
    D.append("samtools index bac_1.validation.bam")
    D.append("python2.7 ~/seqdev/Anchor/tests/reference_validation.py bac_1.anchor.results.txt bac_1 bac_1.validation.bam > bac_1.validation.out")
    for i in range(1,5):
        D.append("python2.7 /home/rdocking/seqdev/Anchor/anchor_pipeline.py -SA -r -f200 -d3 -n25 -c genesis bac_{0}.anchor.fa HS0674.in bac_{1}".format(i, i+1))
    D.submit()
    
    return    

def results_three(root_dir):
    """
    Results 3 - Assemblathon 1
    """

    #   - Create directories for ABySS, SOAP, ALLPATHS-LG
    #   - Submit Anchor jobs
    #   - Submit validation jobs
    
    # ABySS
    C = Cluster_job(jobname = 'anchor_MS_3a.sh', directory = root_dir,
                    mem_token = "4G", mem_free = "4G", h_vmem = "4G")

    C.append("mkdir Assemblathon_ABySS")
    C.append("cd Assemblathon_ABySS")
    C.append("ln -s /scratch/rdocking_prj/Anchor_MS/input_assemblies/speciesA-scaffolds.fa")
    C.append("ln -s /scratch/rdocking_prj/Anchor_MS/input_assemblies/speciesA.fq.in")
    C.append("python2.7 /home/rdocking/seqdev/Anchor/anchor_pipeline.py -r -t100 -f200 -d3 -n25 -c genesis -SA speciesA-scaffolds.fa speciesA.fq.in abyss_1")
    C.append("bwa bwasw /scratch/rdocking_prj/Anchor_MS/reference_sequences/speciesA.hap_1.fa abyss_1.validation.fa | samtools calmd -bSu - /scratch/rdocking_prj/Anchor_MS/reference_sequences/speciesA.hap_1.fa | samtools sort - abyss_1.validation.hap1")    
    C.append("bwa bwasw /scratch/rdocking_prj/Anchor_MS/reference_sequences/speciesA.hap_2.fa abyss_1.validation.fa | samtools calmd -bSu - /scratch/rdocking_prj/Anchor_MS/reference_sequences/speciesA.hap_2.fa | samtools sort - abyss_1.validation.hap2")    
    C.append("samtools index abyss_1.validation.hap1.bam")
    C.append("samtools index abyss_1.validation.hap2.bam")
    C.append("python2.7 ~/seqdev/Anchor/tests/reference_validation.py abyss_1.anchor.results.txt abyss_1 abyss_1.validation.hap1.bam abyss_1.validation.hap2.bam > abyss_1.validation.out")
    for i in range(1,5):
        C.append("python2.7 /home/rdocking/seqdev/Anchor/anchor_pipeline.py -SA -r -f200 -d3 -n25 -c genesis abyss_{0}.anchor.fa speciesA.fq.in abyss_{1}".format(i, i+1))
    C.submit()
    
    # SOAP denovo
    D = Cluster_job(jobname = 'anchor_MS_3b.sh', directory = root_dir,
                    mem_token = "4G", mem_free = "4G", h_vmem = "4G")

    D.append("mkdir Assemblathon_SOAP")
    D.append("cd Assemblathon_SOAP")
    D.append("ln -s /scratch/rdocking_prj/Anchor_MS/input_assemblies/SOAPdenovo-scaffolds.fa")
    D.append("ln -s /scratch/rdocking_prj/Anchor_MS/input_assemblies/speciesA.fq.in")
    D.append("python2.7 /home/rdocking/seqdev/Anchor/anchor_pipeline.py -r -t100 -f200 -d3 -n25 -c genesis -SA SOAPdenovo-scaffolds.fa speciesA.fq.in soap_1")
    D.append("bwa bwasw /scratch/rdocking_prj/Anchor_MS/reference_sequences/speciesA.hap_1.fa soap_1.validation.fa | samtools calmd -bSu - /scratch/rdocking_prj/Anchor_MS/reference_sequences/speciesA.hap_1.fa | samtools sort - soap_1.validation.hap1")    
    D.append("bwa bwasw /scratch/rdocking_prj/Anchor_MS/reference_sequences/speciesA.hap_2.fa soap_1.validation.fa | samtools calmd -bSu - /scratch/rdocking_prj/Anchor_MS/reference_sequences/speciesA.hap_2.fa | samtools sort - soap_1.validation.hap2")    
    D.append("samtools index soap_1.validation.hap1.bam")
    D.append("samtools index soap_1.validation.hap2.bam")
    D.append("python2.7 ~/seqdev/Anchor/tests/reference_validation.py soap_1.anchor.results.txt soap_1 soap_1.validation.hap1.bam soap_1.validation.hap2.bam > test_1.validation.out")
    for i in range(1,5):
        D.append("python2.7 /home/rdocking/seqdev/Anchor/anchor_pipeline.py -SA -r -f200 -d3 -n25 -c genesis soap_{0}.anchor.fa speciesA.fq.in soap_{1}".format(i, i+1))
    D.submit()

    # ALLPATHS-LG
    E = Cluster_job(jobname = 'anchor_MS_3c.sh', directory = root_dir,
                    mem_token = "4G", mem_free = "4G", h_vmem = "4G")

    E.append("mkdir Assemblathon_ALLPATHS")
    E.append("cd Assemblathon_ALLPATHS")
    E.append("ln -s /scratch/rdocking_prj/Anchor_MS/input_assemblies/ALLPATHS-LG-scaffolds.fa")
    E.append("ln -s /scratch/rdocking_prj/Anchor_MS/input_assemblies/speciesA.fq.in")
    E.append("python2.7 /home/rdocking/seqdev/Anchor/anchor_pipeline.py -r -t100 -f200 -d3 -n25 -c genesis -SA ALLPATHS-LG-scaffolds.fa speciesA.fq.in allpaths_1")
    E.append("bwa bwasw /scratch/rdocking_prj/Anchor_MS/reference_sequences/speciesA.hap_1.fa allpaths_1.validation.fa | samtools calmd -bSu - /scratch/rdocking_prj/Anchor_MS/reference_sequences/speciesA.hap_1.fa | samtools sort - allpaths_1.validation.hap1")    
    E.append("bwa bwasw /scratch/rdocking_prj/Anchor_MS/reference_sequences/speciesA.hap_2.fa allpaths_1.validation.fa | samtools calmd -bSu - /scratch/rdocking_prj/Anchor_MS/reference_sequences/speciesA.hap_2.fa | samtools sort - allpaths_1.validation.hap2")    
    E.append("samtools index allpaths_1.validation.hap1.bam")
    E.append("samtools index allpaths_1.validation.hap2.bam")
    E.append("python2.7 ~/seqdev/Anchor/tests/reference_validation.py allpaths_1.anchor.results.txt allpaths_1 allpaths_1.validation.hap1.bam allpaths_1.validation.hap2.bam > allpaths_1.validation.out")
    for i in range(1,5):
        E.append("python2.7 /home/rdocking/seqdev/Anchor/anchor_pipeline.py -SA -r -f200 -d3 -n25 -c genesis allpaths_{0}.anchor.fa speciesA.fq.in allpaths_{1}".format(i, i+1))
    E.submit()

    return

def main():
    """
    Run all Anchor and Anchor-validation jobs associated with the Anchor 
    MS submission
    
    This script requires a lot of the assemblies to be in place already,
    and relies fairly heavily on GSC-specific paths and cluster setup
    
    Usage:  python2.7 anchor_MS_analysis.py main_output_dir
    
    """

    # Starting directory
    root_dir = os.getcwd() + '/' + sys.argv[1]
    os.mkdir(root_dir)    

    # Results 1:    Test vs. known data-set
    results_one(root_dir)
        
    # Results 2:    Small genomes
    results_two(root_dir)
    
    # Results 3:    Assemblathon 1
    results_three(root_dir)
    
    # Results 4:    Big genomes
    #   - Create directories
    #   - Submit Anchor jobs
    
    


if __name__ == '__main__':
    main()

